from .base import BaseProtocol, BaseHoneypot

from struct import unpack

class PostgresProtocol(BaseProtocol):
    '''based on https://github.com/qeeqbox/honeypots/blob/main/honeypots/postgres_server.py'''
    
    config = {
    }

    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "postgres"

    def connection_made(self, transport):
        self.transport = transport
        self._state = 1
        self._variables = {}
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)
        
    def data_received(self, data):    
        
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)

        if self._state == 1:
            if data == b'\x00\x00\x00\x08\x04\xd2\x16/': # SSLRequest
                self._state = 2
                self.transport.write(b'N')
            else:
                # error code for nmap scan
                self.transport.write(b"E\x00\x00\x00\x8b\x53\x46\x41\x54\x41\x4c\x00\x56\x46\x41\x54\x41\x4c\x00\x43\x30\x41\x30\x30\x30\x00\x4d\x75\x6e\x73\x75\x70\x70\x6f\x72\x74\x65\x64\x20\x66\x72\x6f\x6e\x74\x65\x6e\x64\x20\x70\x72\x6f\x74\x6f\x63\x6f\x6c\x20\x36\x35\x33\x36\x33\x2e\x31\x39\x37\x37\x38\x3a\x20\x73\x65\x72\x76\x65\x72\x20\x73\x75\x70\x70\x6f\x72\x74\x73\x20\x33\x2e\x30\x20\x74\x6f\x20\x33\x2e\x30\x00\x46\x70\x6f\x73\x74\x6d\x61\x73\x74\x65\x72\x2e\x63\x00\x4c\x32\x31\x39\x35\x00\x52\x50\x72\x6f\x63\x65\x73\x73\x53\x74\x61\x72\x74\x75\x70\x50\x61\x63\x6b\x65\x74\x00\x00")
                self.transport.close()

        elif self._state == 2:
            self.read_data_custom(data)        
            self._state = 3
            # from https://www.postgresql.org/docs/current/protocol-message-formats.html
            self.transport.write(b'R\x00\x00\x00\x08\x00\x00\x00\x03') #AuthenticationCleartextPassword
            
        elif self._state == 3:
            if data[0] == 112 and 'user' in self._variables:
                self.read_password_custom(data)
                username = self.check_bytes(self._variables['user'])
                password = self.check_bytes(self._variables['password'])
                
                self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={"username": username, "password": password})

                # based on https://www.postgresql.org/docs/current/protocol-message-formats.html
                # ErrorResponse
                message = b'Mpassword authentication failed for user "' + username.encode() + b'"\x00'
                length = ( 4 + 21 + len(message) + 27 ).to_bytes(4, byteorder='big')
                self.transport.write(b'E' + length + b'SFATAL\x00' + b'VFATAL\x00' +b'C28P01\x00' + message + b'\x46\x61\x75\x74\x68\x2e\x63\x00\x4c\x33\x32\x36\x00\x52\x61\x75\x74\x68\x5f\x66\x61\x69\x6c\x65\x64\x00\x00')
                self.transport.close()
            else:
                self.transport.close()

    def check_bytes(self, string):
        if isinstance(string, bytes):
            return string.decode()
        else:
            return str(string)

    def read_data_custom(self, data):
        _data = data.decode('utf-8')
        length = unpack('!I', data[0:4])
        encoded_list = (_data[8:-1].split('\x00'))
        self._variables = dict(zip(*([iter(encoded_list)] * 2)))

    def read_password_custom(self, data):
        data = data.decode('utf-8')
        self._variables['password'] = data[5:].split('\x00')[0]


class PostgresHoneypot(BaseHoneypot):

    def __init__(self, config, logger, bindaddr='0.0.0.0'):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: PostgresProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config



from .http import HttpHandler, HttpHoneypot

from aiohttp import web

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography import x509
from cryptography.x509.oid import NameOID

from pathlib import Path
import ssl
import datetime

class HttpsHandler(HttpHandler):
    def __init__(self, config=None, logger=None):
        super().__init__(config, logger)
        self.protocol_name = "https"

class HttpsHoneypot(HttpHoneypot):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = HttpsHandler(config=config, logger=logger)

        self.COUNTRY_NAME = config.get("country_name") or None
        self.STATE_OR_PROVINCE_NAME = config.get("state_or_province_name") or None
        self.LOCALITY_NAME = config.get("locality_name") or None
        self.ORGANIZATION_NAME = config.get("organization_name") or None
        self.COMMON_NAME = config.get("common_name", "server.internal")
        
        self.key_path = Path(config.get("key", "/etc/trapster/ssl/key.pem"))
        self.certificate_path = Path(config.get("certificate", "/etc/trapster/ssl/certificate.pem"))

        self.generate_certificate()
    
    async def start(self):
        ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(certfile=self.certificate_path, keyfile=self.key_path)

        self.handler.setup()
        app = web.Application()
        app.add_routes([web.route('*', '/{path:.*}', self.handler.handle_request)])
        runner = web.AppRunner(app, access_log=None, handle_signals=True)
        await runner.setup()
        self.site = web.TCPSite(runner, self.bindaddr, self.port, ssl_context=ssl_context)
        await self.site.start()

    def generate_certificate(self):
        '''
        Regenerate the certificate at each startup to ensure the configuration values are applied and reflected.
        '''
        #if self.certificate_path.exists() and self.key_path.exists():
        #    return
        #else:
        self.key_path.parent.mkdir(parents=True, exist_ok=True)
        self.certificate_path.parent.mkdir(parents=True, exist_ok=True)

        key = rsa.generate_private_key(
                public_exponent=65537,
                key_size=2048,
            )

        with open(self.key_path, "wb") as f:
            f.write(key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()        
        ))

        name_attributes = [
            x509.NameAttribute(NameOID.COUNTRY_NAME, self.COUNTRY_NAME) if self.COUNTRY_NAME else None,
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, self.STATE_OR_PROVINCE_NAME) if self.STATE_OR_PROVINCE_NAME else None,
            x509.NameAttribute(NameOID.LOCALITY_NAME, self.LOCALITY_NAME) if self.LOCALITY_NAME else None,
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, self.ORGANIZATION_NAME) if self.ORGANIZATION_NAME else None,
            x509.NameAttribute(NameOID.COMMON_NAME, self.COMMON_NAME),
        ]
        subject = issuer = x509.Name(filter(None, name_attributes))

        alt_names = x509.SubjectAlternativeName([x509.DNSName('localhost'),])

        certification = (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.now())
            .not_valid_after(datetime.datetime.now() + datetime.timedelta(days=90))
            .add_extension(alt_names, False)
            .sign(key, hashes.SHA256(), default_backend())
        )

        with open(self.certificate_path, "wb") as f:
            f.write(certification.public_bytes(serialization.Encoding.PEM))




import requests

def make_query(role, prompt):
    # very basic AI function

    URL = "https://llama-3-8b-instruct.endpoints.kepler.ai.cloud.ovh.net/api/openai_compat/v1/chat/completions"
    OVH_API_KEY = ""
    
    headers = {
        "Authorization": f"Bearer {OVH_API_KEY}",
        "Content-Type": "application/json"
    }

    payload = {
    "max_tokens": 512,
    "messages": [
        {
            "content": "Act as a honeypot server, with realistic responses. Only respond with the corresponding data, and nothing else. Do not write explanations.",
            "name": "system",
            "role": "system"
        },
        {
            "content": prompt,
            "name": role,
            "role": role
        }
    ],
    "model": "Meta-Llama-3-8B-Instruct",
    "temperature": 0,
    }

    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {OVH_API_KEY}",
    }
    
    response = requests.post(URL, json=payload, headers=headers)
    
    if response.status_code == 200:
        response_data = response.json()    
        choices = response_data["choices"]

        for choice in choices:
            text = choice["message"]["content"]
            return text
    else:
        print("Error:", response.status_code)



import asyncio
import psutil
import argparse, json, socket, os
import logging

from . import __version__
from .modules import *
from .logger import set_logger

class TrapsterManager:
    def __init__(self, config):
        self.logger = None
        self.config = config

    def get_ip(self, config_interface):
        for interface, addrs in psutil.net_if_addrs().items():
            if interface == config_interface:
                for addr in addrs:
                    if addr.family == socket.AF_INET:
                        logging.info(f"Using IP: {addr.address} ({config_interface})")
                        return addr.address
        
        logging.warning(f"Interface {config_interface} does not exist, using 0.0.0.0")
        return

    async def start(self):
        ip = self.get_ip(self.config.get('interface', None))
        
        for service_type in self.config['services']:
            for service_config in self.config['services'][service_type]:
                if service_type == 'ftp':
                    server = FtpHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'http':
                    server = HttpHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'https':
                    server = HttpsHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'ssh':
                    server = SshHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'dns':            
                    server = DnsHoneypot(service_config, self.logger, bindaddr=ip, proxy_dns_ip="127.0.0.1")
                elif service_type == 'vnc':
                    server = VncHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'mysql':
                    server = MysqlHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'mssql':
                    server = MssqlHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'postgres':
                    server = PostgresHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'ldap':
                    server = LdapHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'rdp':
                    server = RdpHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'telnet':
                    server = TelnetHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'snmp':
                    server = SnmpHoneypot(service_config, self.logger, bindaddr=ip)
                else:
                    logging.error(f"Unrecognized service {service_type}")
                    break
                                        
                try:
                    logging.info(f"Starting service {service_type} on port {service_config['port']}")
                    await server.start()
                except Exception as e:
                    logging.error(f"Error starting {service_type}: {e}")
        
        while True:
            await asyncio.sleep(10)

def list_interfaces():
    interfaces = psutil.net_if_addrs()
    for interface, addrs in interfaces.items():
        for addr in addrs:
            if addr.family == socket.AF_INET:
                logging.info(f"{addr.address}\t({interface})")

def load_config(config_path):
    if config_path:
        logging.info(f"Using config file at: {config_path}")
    else:
        config_path = os.path.join(os.path.dirname(__file__), "data", "trapster.conf")
        logging.info(f"Using default config file at: {config_path}")
    
    if not os.path.exists(config_path):
        logging.error(f'Config file {config_path} not found')
        return None
    
    with open(config_path, 'r') as f:
        return json.load(f)
    
def main():
    # set logging level to INFO
    logging.basicConfig(level=logging.INFO)
    
    parser = argparse.ArgumentParser(description="Trapster Community honeypot.")
    parser.add_argument('-i', '--interfaces', action='store_true', help='Show list of interfaces and their corresponding IPs.')
    parser.add_argument('-c', '--config', type=str, help='Specify the config file to use.')
    parser.add_argument('-s', '--show-config', action='store_true', help='Show the config file currently in use.')
    parser.add_argument('-v', '--version', action='store_true', help='Print version')
    args = parser.parse_args()
    
    if args.version:
        logging.info(__version__)
        return

    if args.interfaces:
        list_interfaces()
        return

    logging.info('=== Starting Trapster Community ===')
    config = load_config(args.config)
    if not config:
        return

    if args.show_config:
        logging.info(config)
        return

    logger = set_logger(config)
    if logger == None:
        return
      
    manager = TrapsterManager(config)
    logger.whitelist_ips = []
    manager.logger = logger

    try:
        asyncio.run(manager.start())
    except KeyboardInterrupt:
        logging.info('Finishing')



from .base import BaseProtocol, BaseHoneypot

import binascii
from typing import Optional

#TODO add SSL support

class VncProtocol(BaseProtocol):
    
    # All possible version:
    versions = {
        "RFB_33" : b'RFB 003.003\n',
        "RFB_37" : b'RFB 003.007\n',
        "RFB_38" : b'RFB 003.008\n',
    }
    
    config = {
        "version": "RFB_38"
    }

    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "vnc"
        self.challenge = None

    def connection_made(self, transport):
        self.transport = transport
        
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)
        
        self.transport.write(self.versions[self.config['version']])
        self.state = 'wait_pversion'
        
    def data_received(self, data):
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)

        if self.state == 'wait_pversion':
            if data == self.versions[self.config['version']]:
                # sending security types
                self.transport.write(binascii.unhexlify('020217'))
                self.state = 2
            else:
                self.unrecognized_data(data)
        elif self.state == 2:
            if data == binascii.unhexlify('02'):
                self.accept_vnc_authentication()
            else:
                self.unrecognized_data(data)
        elif self.state == 3:
            self.authentication(data)

    def make_challenge(self):
        from os import urandom
        self.challenge = binascii.hexlify(urandom(16))        
        return binascii.unhexlify(self.challenge)

    def accept_vnc_authentication(self):
        self.transport.write(self.make_challenge())
        self.state = 3

    def authentication(self, data):
        # needs to be truncated to 8bytes length (max password length)
        response = binascii.hexlify(data)[:16]
        challenge = self.challenge[:16]


        # https://hashcat.net/forum/thread-8833-post-46908.html
        # Clear password can be retrieved using hashcat (very long)
        # print("./hashcat.bin -m 14000 -a 3  -1 resources/VNC_ascii.hcchr --hex-charset {}:{} '?1?1?1?1?1?1?1?1'".format(
        #    response.decode(),
        #    challenge.decode()
        #    ))
        # 
        # VNC_ascii.hcchr (all 95 ascii characters transfomed for VNC) with 00:
        # 8646c626a666e6169656d636b676f60e8e4ece2eae6eee1e9e5e8242c222a262e2129252d232b272f20a8a4aca2aaa6aea1a9a5a0c8c4ccc2cac6cec1c9c840224a4547ab4d4fabc7edabadebe5cdc3c7c3474fcf43a449414e46406043e00

        self.transport.write(binascii.unhexlify('00000001')) #response code
        self.transport.write(binascii.unhexlify('00000016')) #message length
        self.transport.write(b'Authentication failure') #message
        self.transport.close()
        
        self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={'response': response.decode(), 'challenge': challenge.decode()})

    def unrecognized_data(self, data):
        self.transport.close()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        self.state = 0
        self.transport.close()

class VncHoneypot(BaseHoneypot):
    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: VncProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config


import struct

# details of dns packet : https://courses.cs.duke.edu/fall16/compsci356/DNS/DNS-primer.pdf
# code from https://stackoverflow.com/questions/16977588/reading-dns-packets-in-python

def decode_labels(message, offset):
    labels = []

    while True:
        length, = struct.unpack_from("!B", message, offset)

        if (length & 0xC0) == 0xC0:
            pointer, = struct.unpack_from("!H", message, offset)
            offset += 2

            return labels + decode_labels(message, pointer & 0x3FFF), offset

        if (length & 0xC0) != 0x00:
            raise "unknown label encoding"

        offset += 1

        if length == 0:
            return labels, offset

        labels.append(*struct.unpack_from("!%ds" % length, message, offset))
        try:
            labels[-1] = labels[-1].decode()
        except UnicodeDecodeError:
            labels[-1] = str(labels[-1])

        offset += length


DNS_QUERY_SECTION_FORMAT = struct.Struct("!2H")

def decode_question_section(message, offset, qdcount):
    questions = []

    for _ in range(qdcount):
        qname, offset = decode_labels(message, offset)

        qtype, qclass = DNS_QUERY_SECTION_FORMAT.unpack_from(message, offset)
        offset += DNS_QUERY_SECTION_FORMAT.size

        question = {"domain_name": qname,
                    "query_type": qtype,
                    "query_class": qclass}

        questions.append(question)

    return questions, offset


DNS_QUERY_MESSAGE_HEADER = struct.Struct("!6H")

def decode_dns_message(message):

    id, flags, qdcount, ancount, nscount, arcount = DNS_QUERY_MESSAGE_HEADER.unpack_from(message)
    
    qr = (flags & 0x8000) != 0 
    opcode = (flags & 0x7800) >> 11 
    aa = (flags & 0x0400) != 0 
    tc = (flags & 0x200) != 0 
    rd = (flags & 0x100) != 0 
    ra = (flags & 0x80) != 0
    z = (flags & 0x70) >> 4
    rcode = flags & 0xF

    offset = DNS_QUERY_MESSAGE_HEADER.size
    questions, offset = decode_question_section(message, offset, qdcount)

    result = {"id": id,
              "is_response": qr,
              "opcode": opcode,
              "is_authoritative": aa,
              "is_truncated": tc,
              "recursion_desired": rd,
              "recursion_available": ra,
              "reserved": z,
              "response_code": rcode,
              "question_count": qdcount,
              "answer_count": ancount,
              "authority_count": nscount,
              "additional_count": arcount,
              "questions": questions}

    return result


from .base import BaseProtocol, BaseHoneypot, UdpTransporter

import asyncio
from scapy.all import SNMP
from scapy.error import Scapy_Exception

class SnmpUdpProtocol(BaseProtocol):
    '''
    SNMP log only server
    TODO: could be improve to respond to messages with custom data
    '''
    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "snmp"
    
    def connection_made(self, transport) -> None:
        self.transport = transport
        
    async def GetRequest():
        return
    async def GetResponse():
        return
    async def SetRequest():
        return

    def datagram_received(self, data, addr):
        version, community, oids = self.parse_snmp(data)
        if version or community or oids:
            src_ip, src_port = addr
            dst_ip, dst_port = self.transport.get_extra_info('sockname')
            transport_udp = UdpTransporter(dst_ip, dst_port, src_ip, src_port)
            self.logger.log(self.protocol_name + "." + self.logger.QUERY, transport_udp, extra={"community":community, "version":version, "varbind":oids})

    def parse_snmp(self, data):
        try:
            snmp_data = SNMP(data)
            try:
                community = snmp_data.community.val.decode(errors='backslashreplace')
            except AttributeError:
                community = snmp_data.community.val
            oids = " ".join([item.oid.val for item in snmp_data.PDU.varbindlist])
            version = snmp_data.version.val

        except Scapy_Exception:
            #TODO: Scapy does not support SNMPv3 by default, but a new class can be created to support it, like in PySNMP
            version = "unknown"
            community = "unknown"
            oids = "unknown"
            
        return version, community, oids

class SnmpHoneypot(BaseHoneypot):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: SnmpUdpProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config

        self.handler_udp = SnmpUdpProtocol

    async def _start_server(self):
        loop = asyncio.get_running_loop()

        # Create UDP server
        transport, protocol = await loop.create_datagram_endpoint(lambda: self.handler_udp(), 
                                    local_addr=(self.bindaddr, self.port))


import pytest
import asyncio
import asyncssh
from trapster.modules.ssh import SshHoneypot
from trapster.logger import JsonLogger

from .base import BaseServerTest


class SSHTest(BaseServerTest):
    async def connect_and_login(self, port, username="invalid_user", password="invalid_pass"):
        try:
            async with asyncssh.connect(self.bindaddr, port, username=username, password=password, known_hosts=None) as conn:
                # If login succeeds, it's a failure in the test
                pytest.fail("Expected authentication to fail, but it succeeded")
        except asyncssh.PermissionDenied:
            print("Authentication failed as expected")
            # Authentication failed as expected; this is the desired outcome
            return True
        except Exception as e:
            pytest.fail(f"Unexpected error occurred: {e}")
            return False

    async def run_test(self):
        await self.connect_and_login(self.service_config['port'])


@pytest.mark.asyncio
async def test_ssh_authentication_failed():
    service_config = {
        'port': 2222,
    }
    logger = JsonLogger('trapster-1')

    # Initialize the SSH-specific test
    ssh_test = SSHTest(SshHoneypot, service_config, logger)

    # Start the server
    await ssh_test.start_server()

    try:
        # Run the SSH-specific test logic
        await ssh_test.run_test()
    finally:
        # Stop the server
        await ssh_test.stop_server()


from aiohttp import web
from jinja2.sandbox import ImmutableSandboxedEnvironment
from jinja2 import FileSystemLoader, Undefined
import yaml
import random, string, base64, mimetypes, re
from datetime import datetime, timezone
from pathlib import Path

from .base import BaseHoneypot
from .libs import ai

class HttpHandler:
    def __init__(self, config, logger):
        self.protocol_name = "http"

        self.logger = logger

        self.NAME = config.get('skin', 'default_apache')
        self.BASIC_AUTH = config.get('basic_auth', False)
        self.USERNAME = config.get('username', None)
        self.PASSWORD = config.get('password', None)

        self.data_folder = Path(__file__).parent.parent / "data" / "http"

    def setup(self):
        try:
            resolved_path = (self.data_folder / self.NAME).resolve()
            if not resolved_path.is_relative_to(self.data_folder):
                raise ValueError(f"Invalid skin name: {self.NAME}")
        except (ValueError, RuntimeError):
            self.NAME = "default_apache"  # Fallback to a default skin

        self.static_folder = self.data_folder / self.NAME / "files"
        self.template_folder = self.data_folder / self.NAME / "templates"
        self.config_file = self.data_folder / self.NAME / "config.yaml"

        with self.config_file.open('r') as file:
            self.http_config = yaml.safe_load(file)
        
        self.env = self.create_jinja_env()

    @staticmethod
    def sanitize_request(request):
        if request:
            return { 
                "url": request.url,
                "path": request.path,
                "method": request.method,
                "headers": dict(request.headers),
                "body": request.content if request.body_exists else None,
                "remote": request.remote, 
                "cookies": request.cookies,
                "query_string": request.query_string,
                "content_type": request.content_type,
                "host": request.host,
                "secure": request.secure,
                "scheme": request.scheme,
                "path_qs": request.path_qs
            }
        
    @staticmethod
    def random_filter(seed=None, alphabet=string.hexdigits[:-6], length=36):
        # Jinja filter to generate a random string
        if seed is not None:
            random.seed(seed)
        return ''.join(random.choice(alphabet) for _ in range(length))

    def create_jinja_env(self):
        env = ImmutableSandboxedEnvironment(
            loader=FileSystemLoader(self.template_folder),
            autoescape=True
        )
        env.globals.update({
            'random': self.random_filter,
            'get_current_time': lambda: datetime.now(timezone.utc).strftime('%a, %d %b %Y %H:%M:%S GMT'),
        })
        env.undefined = Undefined
        return env

    def get_endpoint_config(self, full_url, method):
        for endpoint in self.http_config.get('endpoints', []):
            for route, details in endpoint.items():
                if re.fullmatch(route, full_url):
                    if isinstance(details, list):
                        return next((d for d in details if d['method'] == method), None)
                    elif details['method'] == method:
                        return details
        return None

    def get_content(self, endpoint_config, request=None):
        if not endpoint_config:
            return "", 200
        
        if 'content' in endpoint_config:
            return endpoint_config['content'], endpoint_config.get('status_code', 200)
        
        elif 'file' in endpoint_config:
            file_path = self.template_folder / endpoint_config['file']
            try:
                if file_path.resolve().relative_to(self.template_folder.resolve()):
                    file_path = file_path.resolve()
                    with file_path.open('r') as file:
                         # add request object to the template
                        template = self.env.from_string(file.read())
                        template.globals['request'] = self.sanitize_request(request)

                    return template.render(), 200
            except (ValueError, FileNotFoundError):
                pass

        elif 'ai' in endpoint_config:
            # experimental ai response
            prompt = endpoint_config['ai']['prompt'].replace("{{ path }}", request.path)
            return ai.make_query('user', prompt), 200
        
        return "File not found", 404

    async def handle_error(self, request, error_code):
        # Check if errors is defined in config.yaml, otherwise use a error template (like 401.html), otherwise send nothing
        error_config = self.http_config.get('errors', {}).get(str(error_code))
        if error_config:
            content, _ = self.get_content(error_config)
        else:
            content, _ = self.get_content({"file": f"{error_code}.html"})

        headers = self.http_config.get('headers', {}).copy()
        headers.update(self.http_config.get('errors', {}).get(str(error_code), {}).get('headers', {}))
        
        if error_code == 401:
            headers['WWW-Authenticate'] = 'Basic realm="Restricted Area"'
        
        await self.log(request, self.logger.QUERY, error_code)
        return web.Response(body=content, content_type='text/html', status=error_code, headers=headers)

    async def handle_request(self, request):
        if self.BASIC_AUTH and not self.check_auth(request):
            return await self.handle_error(request, 401)

        full_url = request.url.path_qs
        method = request.method
        endpoint_config = self.get_endpoint_config(full_url, method)
        headers = {}

        if endpoint_config:
            # Use configured response    
            content, status_code = self.get_content(endpoint_config, request)
            status_code = endpoint_config.get('status_code', status_code)
            headers = endpoint_config.get('headers', {})
            await self.log(request, self.logger.QUERY, status_code)
        else:
            content, status_code, headers = await self.handle_static_file(request)
            # only default response is logged

        # Prepare response headers
        response_headers = self.http_config.get('headers', {}).copy()
        response_headers.update(headers)

        # Determine content type
        content_type = response_headers.pop('Content-Type', 'text/html')

        return web.Response(body=content, content_type=content_type, charset='utf-8', 
                            status=status_code, headers=response_headers)

    def check_auth(self, request):
        auth_header = request.headers.get('Authorization')
        if not auth_header or not auth_header.startswith('Basic '):
            return False
        encoded_credentials = auth_header.split(' ', 1)[1]
        username, password = base64.b64decode(encoded_credentials).decode('utf-8').split(':')
        self.logger.log(f"{self.protocol_name}.{self.logger.LOGIN}", request.transport, 
                        extra={"username": username, "password": password})
        return username == self.USERNAME and password == self.PASSWORD

    async def handle_static_file(self, request):
        if request.path.endswith('/'):
            file_path = self.static_folder / request.path.lstrip('/') / 'index.html'
        elif request.method == 'GET':
            file_path = self.static_folder / request.path.lstrip('/')
        else:
            return await self.handle_default(request)

        try:
            if file_path.resolve().relative_to(self.static_folder) and file_path.is_file():
                content = file_path.read_bytes()
                content_type = mimetypes.guess_type(str(file_path))[0] or 'application/octet-stream'
                return content, 200, {'Content-Type': content_type}
        except ValueError:
            return "", 500, {}

        return await self.handle_default(request)

    async def handle_default(self, request):
        default_response = self.http_config.get('default')
        content, _ = self.get_content(default_response)
        status_code = default_response.get('status_code', 404)
        headers = default_response.get('headers', {})
        await self.log(request, self.logger.QUERY, status_code)
        return content, status_code, headers
    
    async def log(self, request, log_type, status_code, extra=None):
        all_extra = {
            "method": request.method,
            "target": request.path_qs,
            "version": f"HTTP/{request.version.major}.{request.version.minor}",
            "headers": dict(request.headers),
            "status_code": status_code,
        }

        if request.body_exists:
            all_extra["data"] = (await request.read()).decode(errors='backslashreplace')
        all_extra.update(extra or {})
        self.logger.log(f"{self.protocol_name}.{log_type}", request.transport, extra=all_extra)


class HttpHoneypot(BaseHoneypot):
    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.port = config['port']
        self.handler = HttpHandler(config=config, logger=logger)

    async def start(self):
        self.handler.setup()
        app = web.Application()
        app.add_routes([web.route('*', '/{path:.*}', self.handler.handle_request)])
        runner = web.AppRunner(app, access_log=None, handle_signals=True)
        await runner.setup()
        self.site = web.TCPSite(runner, self.bindaddr, self.port)
        await self.site.start()

    async def stop(self):
        if hasattr(self, 'site'):
            await self.site.stop()



from .base import BaseProtocol, BaseHoneypot

import re

class RdpProtocol(BaseProtocol):

    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "rdp"
        self.initial_connection = True

    def connection_made(self, transport) -> None:
        self.transport = transport
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)
       

    def data_received(self, data):
      
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)

        decoded_data = data.decode(encoding="utf-8", errors="ignore")
        # Use regex to extract the username.
        match = re.search(r"mstshash=(?P<username>[a-zA-Z0-9-_@]*)", decoded_data)
        username = match and match.groupdict().get("username")

        if self.initial_connection:
            self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={"username": username, "password": ''})

            # Respond as an NLA-enabled RDP server
            self.transport.write(
                b"\x03\x00\x00\x13\x0e\xd0\x00\x00\x124\x00\x02\x1f\x08\x00\x08\x00\x00\x00"
            )
            self.initial_connection = False
        else:
            # Always respond with a negotiation failure, details from
            # https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/96327ab4-d43f-4803-9aff-392ce1fc2073
            self.transport.write(bytes.fromhex("0001000400010000052e"))
            self.connection_lost('')

class RdpHoneypot(BaseHoneypot):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: RdpProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config



from .vnc import VncHoneypot
from .http import HttpHoneypot
from .https import HttpsHoneypot
from .ssh import SshHoneypot
from .ftp import FtpHoneypot
from .mysql import MysqlHoneypot
from .telnet import TelnetHoneypot
from .dns import DnsHoneypot
from .rdp import RdpHoneypot
from .postgres import PostgresHoneypot
from .ldap import LdapHoneypot
from .snmp import SnmpHoneypot
from .mssql import MssqlHoneypot